from .aaware_whisper import aaware_whisper
from .aixplain_whisper import aixplain_whisper
from .deepgram import deepgram
from .fastwhisper import fastwhisper
from .google import google
from .whisper import whisper
